
#include "include/types.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/riscv.h"
#include "include/spinlock.h"
#include "include/sleeplock.h"
#include "include/proc.h"
#include "include/elf.h"
#include "include/fat32.h"
#include "include/kalloc.h"
#include "include/vm.h"
#include "include/printf.h"
#include "include/string.h"

//read and check elf header
//return 0 on success,-1 on failure
static int
readelfhdr(struct dirent* ep,struct elfhdr* elf){
  // Check ELF header
  if(eread(ep, 0, (uint64)elf, 0, sizeof(struct elfhdr)) != sizeof(struct elfhdr))
    return -1;
  if(elf->magic != ELF_MAGIC)
    return -1;
  return 0;
}

// Load a program segment into pagetable at virtual address va.
// va must be page-aligned
// and the pages from va to va+sz must already be mapped.
// Returns 0 on success, -1 on failure.
static int
loadseg(pagetable_t pagetable, uint64 va, struct dirent *ep, uint offset, uint sz)
{
  uint i, n;
  uint64 pa;
  i = 0;
  if((va % PGSIZE) != 0){
    //printf("not aligned va:%p\n",va);
    uint64 off = va % PGSIZE;
    uint64 rest = PGSIZE -off;
    pa = walkaddr(pagetable, va)+off;
    //printf("not aligned pa:%p\n",pa);
    if(pa == NULL)
      panic("loadseg: address should exist");
    if(sz - i < rest)
      n = sz - i;
    else
      n = rest;
    //printf("map zone:[%p,%p]\n",va,va+n);
    if(eread(ep, 0, (uint64)pa, offset+i, n) != n)
      return -1;
    i += n;
  }

  // pa = walkaddr(pagetable, va);
  // printf("va->%p\tpa->%p\tsize->%d\n", va, pa, sz);
  
  for(; i < sz; i += PGSIZE){
    pa = walkaddr(pagetable, va + i);
    if(pa == NULL)
      panic("loadseg: address should exist");
    if(sz - i < PGSIZE)
      n = sz - i;
    else
      n = PGSIZE;
    if(eread(ep, 0, (uint64)pa, offset+i, n) != n)
      return -1;
  }

  return 0;
}

uint64 
linkermap(uint64 va){
  return LINKER_OFFSET+va;
}

void*
checklinker(pagetable_t pagetable,uint64 va){
  uint64 linkerva = linkermap(va);
  uint64* linkerpa = (void*)kwalkaddr1(pagetable,linkerva);
  printf("dynamic linker %p map to %p\n",linkerva,linkerpa);
  if(linkerpa)printf("instruction:%p\n",*(uint64*)linkerpa);
  return linkerpa;
}

void*
checkapp(pagetable_t pagetable,uint64 va){
  uint64* pa = (void*)kwalkaddr1(pagetable,va);
  printf("dynamic linker %p map to %p\n",va,pa);
  if(pa)printf("instruction:%p\n",*(uint64*)pa);
  return pa;
}

static void
auxalloc(uint64* aux,uint64 atid,uint64 value)
{
  //printf("aux[%d] = %p\n",atid,value);
  uint64 argc = aux[0];
  aux[argc*2+1] = atid;
  aux[argc*2+2] = value;
  aux[argc*2+3] = 0;
  aux[argc*2+4] = 0;
  aux[0]++;
}

// Load a program elf into pagetable
// Returns sz on success, -1 on failure.
// ep must be locked.
uint64
loadlinker(pagetable_t pagetable,pagetable_t kpagetable, struct dirent *ep,struct elfhdr* elf,
		struct elfhdr* appelf,struct proghdr* appphdr,uint64* entry,uint64 sz,uint64* aux)
{
  //printf("MAXVA:%p\n",MAXVA);
  struct proghdr ph;
  auxalloc(aux,AT_BASE ,LINKER_OFFSET);
  if(appphdr->type != 0xffffffff){
    //printf("app phdr:%p\n",appphdr->vaddr);
    auxalloc(aux,AT_PHDR,appphdr->vaddr);
  }
  else return -1;
  auxalloc(aux,AT_UID,0);
  auxalloc(aux,AT_EUID,0);
  auxalloc(aux,AT_GID,0);
  auxalloc(aux,AT_EGID,0);
  auxalloc(aux,AT_SECURE,0);
  auxalloc(aux,AT_ENTRY,appelf->entry);
  auxalloc(aux,AT_PHNUM,appelf->phnum);
  auxalloc(aux,AT_PHENT,appelf->phentsize);
  auxalloc(aux,AT_PAGESZ,PGSIZE);
  for(int i=0, off=elf->phoff; i<elf->phnum; i++, off+=sizeof(ph)){
    if(eread(ep, 0, (uint64)&ph, off, sizeof(ph)) != sizeof(ph))
      return -1;
    if(ph.type == ELF_PROG_LOAD){
      if(ph.memsz < ph.filesz)
        return -1;
      uint64 linkerva = linkermap(ph.vaddr);
      if(linkerva + ph.memsz < linkerva)
        return -1;
      uint64 sz1;
      if((sz1 = uvmalloc(pagetable, kpagetable, sz, PGROUNDUP(linkerva) + ph.memsz)) == 0){
        return -1;
      }
      sz = sz1;
      /*
      if(ph.vaddr % PGSIZE != 0){
        printf("ph.vaddr not aligned:%p\n",ph.vaddr);
        goto bad;
      }
      */
      if(loadseg(pagetable, linkerva, ep, ph.off, ph.filesz) < 0)
        return -1;
    }
  }
  /*
  if(loadsec(pagetable,ep,elf)<0){
    return -1;
  }
  */
  
  //uint16* linkerpa = checklinker(pagetable,0x53066);
  //checklinker(pagetable,0x8d3b8);
  //*linkerpa = 0;
  
  return sz;
}

// Load a program elf into pagetable
// Returns sz on success, -1 on failure.
// ep must be locked.
static uint64
loadelf(pagetable_t pagetable,pagetable_t kpagetable, struct dirent *ep,struct elfhdr* elf,uint64* entry,uint64 sz,uint64* aux)
{
  struct proghdr ph;
  struct proghdr phdr = {0xffffffff};
  struct elfhdr  linkelf;
  for(int i=0, off=elf->phoff; i<elf->phnum; i++, off+=sizeof(ph)){
    if(eread(ep, 0, (uint64)&ph, off, sizeof(ph)) != sizeof(ph))
      return -1;
    if(ph.type == ELF_PROG_LOAD){
      if(ph.memsz < ph.filesz)
        return -1;
      if(ph.vaddr + ph.memsz < ph.vaddr)
        return -1;
      uint64 sz1;
      if((sz1 = uvmalloc(pagetable, kpagetable, sz, PGROUNDUP(ph.vaddr) + ph.memsz)) == 0)
        return -1;
      sz = sz1;
      /*
      if(ph.vaddr % PGSIZE != 0){
        printf("ph.vaddr not aligned:%p\n",ph.vaddr);
        goto bad;
      }
      */
      if(loadseg(pagetable, ph.vaddr, ep, ph.off, ph.filesz) < 0)
        return -1;
    }else if(ph.type==ELF_PROG_PHDR){
      phdr = ph;
    }else if(ph.type==ELF_PROG_INTERP){
      char mybuf[50];
      if(eread(ep, 0, (uint64)mybuf, ph.off,ph.filesz)!=ph.filesz)
        return -1;
      
      if(strncmp(mybuf,"/lib/ld-musl-riscv64.so.1",ph.filesz)==0 || strncmp(mybuf,"/lib/ld-musl-riscv64-sf.so.1",ph.filesz) == 0){
        strncpy(mybuf,"libc.so",50);
      }
      struct dirent* linkep;
      if((linkep=ename(mybuf))==NULL){
        printf("[exec]linker %s not found\n",mybuf);
        return -1;
      }
      elock(linkep);
      if(readelfhdr(linkep,&linkelf)<0){
        return -1;
      }
      //printf("dynamic linker entry:%p\n",linkelf.entry);
      
      if((sz = loadlinker(pagetable,kpagetable,linkep,&linkelf,elf,&phdr,entry,sz,aux)) == -1){
        printf("[exec]linker sz not valid\n");
        return -1;
      }
      
      //printf("dynamic linker entry:%p map to %p\n",linkelf.entry,kwalkaddr1(pagetable,linkelf.entry));
      *entry = linkermap(linkelf.entry);
      eunlock(linkep);
      eput(linkep);
    }
  }
  return sz;
}

static uint64
loadaux(pagetable_t pagetable,uint64 sp,uint64 stackbase,uint64* aux){
  int argc = aux[0];
  if(!argc)return sp;
  /*
  printf("aux argc:%d\n",argc);
  for(int i=1;i<=2*argc+2;i++){
    printf("final raw aux[%d] = %p\n",i,aux[i]);
  }
  */
  sp -= (2*argc+2) * sizeof(uint64);
  if(sp < stackbase){
    return -1;
  }
  aux[0] = 0;
  if(copyout(pagetable, sp, (char *)(aux+1), (2*argc+2)*sizeof(uint64)) < 0){
    return -1;
  }
  return sp;
}

static uint64 
ustackpushstr(pagetable_t pagetable,uint64* ustack,char* str,uint64 sp,uint64 stackbase)
{
  uint64 argc = ++ustack[0];
  if(argc>MAXARG+1)return -1;
  sp -= strlen(str) + 1;
  sp -= sp % 16; // riscv sp must be 16-byte aligned
  if(sp < stackbase)
      return -1;
  if(copyout(pagetable, sp, str, strlen(str) + 1) < 0)
      return -1;
  ustack[argc] = sp;
  ustack[argc+1] = 0;
  return sp;
}

void
stackdisplay(pagetable_t pagetable,uint64 sp,uint64 sz)
{
  for(uint64 i = sp;i<sz;i+=8){
    uint64 *pa = (void*)kwalkaddr1(pagetable,i);
    if(pa)printf("addr %p value %p\n",i,*pa);
    else printf("addr %p value (nil)\n",i);
  }
}

int exec(char *path, char **argv)
{
  uint64 aux[AUX_CNT*2+3] = {0,0,0};
  uint64 env[10]={0};
  char *s, *last;
  uint64 argc, sz = 0, sp, ustack[MAXARG+2], stackbase,progentry;
  struct elfhdr elf;
  struct dirent *ep;
  pagetable_t pagetable = 0, oldpagetable;
  pagetable_t kpagetable = 0, oldkpagetable;
  struct proc *p = myproc();
  // Make a copy of p->kpt without old user space, 
  // but with the same kstack we are using now, which can't be changed
  if ((kpagetable = (pagetable_t)kalloc()) == NULL) {
    return -1;
  }
  memmove(kpagetable, p->kpagetable, PGSIZE);

  for (int i = 0; i < PX(2, MAXUVA); i++) {
    kpagetable[i] = 0;
  }


  if((ep = ename(path)) == NULL) {
    #ifdef DEBUG
    printf("[exec] %s not found\n", path);
    #endif
    goto bad;
  }
  elock(ep);
  // Check ELF header
  if(readelfhdr(ep,&elf)<0){
    goto bad;
  }
  progentry = elf.entry;
  if((pagetable = proc_pagetable(p)) == NULL)
    goto bad;
  /*
  printf("elf prohdr:\n");
  for(i=0, off=elf.phoff; i<elf.phnum; i++, off+=sizeof(ph)){
    if(eread(ep, 0, (uint64)&ph, off, sizeof(ph)) != sizeof(ph))
      goto bad;
    printf("type:%d\toffset:%p\tfilesz:%p\tvaddr:%p\tpaddr:%p\n",ph.type,ph.off,ph.filesz,ph.vaddr,ph.paddr);
  }
  */
  // Load program into memory.
  sz = loadelf(pagetable,kpagetable,ep,&elf,&progentry,0,aux);
  eunlock(ep);
  eput(ep);
  ep = 0;
  if(sz==-1){
    printf("[exec]sz not valid\n");
    goto bad;
  }
  p = myproc();
  uint64 oldsz = p->sz;

  // Allocate two pages at the next page boundary.
  // Use the second as the user stack.
  sz = PGROUNDUP(sz);
  uint64 sz1;
  if((sz1 = uvmalloc(pagetable, kpagetable, sz, sz + 2*PGSIZE)) == 0)
    goto bad;
  sz = sz1;
  uvmclear(pagetable, sz-2*PGSIZE);
  sp = sz;
  stackbase = sp - PGSIZE;
  
  ustack[0]=0;
  //entry in linker
  if(progentry!=elf.entry){
  
    //("elf.entry:%p\n",elf.entry);
    //printf("progentry:%p\n",progentry);
    /*
    if((sp = ustackpushstr(pagetable,ustack,"./libc.so",sp,stackbase))==-1){
      goto bad;
    }
    */

    if((sp = ustackpushstr(pagetable,env,"LD_LIBRARY_PATH=/",sp,stackbase))==-1){
      goto bad;
    }
  }
  
  // Push argument strings, prepare rest of stack in ustack.
  for(argc = 0; argv[argc]; argc++) {
    if((sp = ustackpushstr(pagetable,ustack,argv[argc],sp,stackbase))==-1){
      goto bad;
    }
  }
  
  if((env[0]+ustack[0]+1)%2){sp -= 8;}//16 aligned
  if((sp = loadaux(pagetable,sp,stackbase,aux))<0){
    printf("[exec]pass aux too many\n");
    goto bad;
  }
  argc = env[0];
  if(argc){
    // push the array of argv[] pointers.
    sp -= (argc+1) * sizeof(uint64);
    if(sp < stackbase){
      goto bad;
    }
    if(copyout(pagetable, sp, (char *)(env+1), (argc+1)*sizeof(uint64)) < 0){
      goto bad;
    }
  }
  argc = ustack[0];
  // push the array of argv[] pointers.
  sp -= (argc+2) * sizeof(uint64);
  if(sp < stackbase){
    goto bad;
  }
  
  if(copyout(pagetable, sp, (char *)ustack, (argc+2)*sizeof(uint64)) < 0){
    goto bad;
  }
  //stackdisplay(pagetable,sp,sz);
  // arguments to user main(argc, argv)
  // argc is returned via the system call return
  // value, which goes in a0.
  p->trapframe->a1 = sp;

  // Save program name for debugging.
  for(last=s=path; *s; s++)
    if(*s == '/')
      last = s+1;
  safestrcpy(p->name, last, sizeof(p->name));

  // printf("ready to execute %s\n", p->name);

  // Commit to the user image.
  oldpagetable = p->pagetable;
  oldkpagetable = p->kpagetable;
  p->pagetable = pagetable;
  p->kpagetable = kpagetable;
  //printf("[exec]old pagetable:%p new pagetable:%p\n",oldpagetable,pagetable);
  p->sz = sz;
  p->trapframe->epc = progentry;  // initial program counter = main
  p->trapframe->sp = sp; // initial stack pointer
  proc_freepagetable(oldpagetable, oldsz);
  w_satp(MAKE_SATP(p->kpagetable));
  sfence_vma();
  kvmfree(oldkpagetable, 0);
  return argc; // this ends up in a0, the first argument to main(argc, argv)

 bad:
  #ifdef DEBUG
  printf("[exec] reach bad\n");
  #endif
  
  if(pagetable)
    proc_freepagetable(pagetable, sz);
  if(kpagetable)
    kvmfree(kpagetable, 0);
  if(ep){
    eunlock(ep);
    eput(ep);
  }
  return -1;
}
