
#include "include/types.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/riscv.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/syscall.h"
#include "include/sysinfo.h"
#include "include/kalloc.h"
#include "include/vm.h"
#include "include/string.h"
#include "include/printf.h"

// Fetch the uint64 at addr from the current process.
int
fetchaddr(uint64 addr, uint64 *ip)
{
  struct proc *p = myproc();
  if(addr >= p->sz || addr+sizeof(uint64) > p->sz)
    return -1;
  // if(copyin(p->pagetable, (char *)ip, addr, sizeof(*ip)) != 0)
  if(copyin2((char *)ip, addr, sizeof(*ip)) != 0)
    return -1;
  return 0;
}

// Fetch the nul-terminated string at addr from the current process.
// Returns length of string, not including nul, or -1 for error.
int
fetchstr(uint64 addr, char *buf, int max)
{
  // struct proc *p = myproc();
  // int err = copyinstr(p->pagetable, buf, addr, max);
  int err = copyinstr2(buf, addr, max);
  if(err < 0)
    return err;
  return strlen(buf);
}

static uint64
argraw(int n)
{
  struct proc *p = myproc();
  switch (n) {
  case 0:
    return p->trapframe->a0;
  case 1:
    return p->trapframe->a1;
  case 2:
    return p->trapframe->a2;
  case 3:
    return p->trapframe->a3;
  case 4:
    return p->trapframe->a4;
  case 5:
    return p->trapframe->a5;
  }
  panic("argraw");
  return -1;
}

// Fetch the nth 32-bit system call argument.
int
argint(int n, int *ip)
{
  *ip = argraw(n);
  return 0;
}

// Retrieve an argument as a pointer.
// Doesn't check for legality, since
// copyin/copyout will do that.
int
argaddr(int n, uint64 *ip)
{
  *ip = argraw(n);
  return 0;
}

// Fetch the nth word-sized system call argument as a null-terminated string.
// Copies into buf, at most max.
// Returns string length if OK (including nul), -1 if error.
int
argstr(int n, char *buf, int max)
{
  uint64 addr;
  if(argaddr(n, &addr) < 0)
    return -1;
  return fetchstr(addr, buf, max);
}

extern uint64 sys_chdir(void);
extern uint64 sys_close(void);
extern uint64 sys_dup(void);
extern uint64 sys_exec(void);
extern uint64 sys_exit(void);
extern uint64 sys_fork(void);
extern uint64 sys_fstat(void);
extern uint64 sys_getpid(void);
extern uint64 sys_kill(void);
extern uint64 sys_mkdir(void);
extern uint64 sys_open(void);
extern uint64 sys_pipe(void);
extern uint64 sys_read(void);
extern uint64 sys_sbrk(void);
extern uint64 sys_sleep(void);
extern uint64 sys_wait(void);
extern uint64 sys_write(void);
extern uint64 sys_uptime(void);
extern uint64 sys_test_proc(void);
extern uint64 sys_dev(void);
extern uint64 sys_readdir(void);
extern uint64 sys_getcwd(void);
extern uint64 sys_remove(void);
extern uint64 sys_trace(void);
extern uint64 sys_sysinfo(void);
extern uint64 sys_rename(void);
extern uint64 sys_sprint(void);
extern uint64 sys_openat(void);
extern uint64 sys_brk(void);
extern uint64 sys_munmap(void);
extern uint64 sys_mmap(void);
extern uint64 sys_getppid(void);
extern uint64 sys_getpid(void);
extern uint64 sys_mkdirat(void);
extern uint64 sys_dup3(void);
extern uint64 sys_clone(void);
extern uint64 sys_wait4(void);
extern uint64 sys_sched_yield(void);
extern uint64 sys_times(void);
extern uint64 sys_nanosleep(void);
extern uint64 sys_gettimeofday(void);
extern uint64 sys_execve(void);
extern uint64 sys_getdents64(void);
extern uint64 sys_uname(void);
extern uint64 sys_unlinkat(void);
extern uint64 sys_umount(void);
extern uint64 sys_mount(void);
extern uint64 sys_finit(void);
extern uint64 sys_exit_group(void);
extern uint64 sys_set_tid_address(void);
extern uint64 sys_rt_sigaction(void);
extern uint64 sys_rt_sigprocmask(void);
extern uint64 sys_rt_sigtimedwait(void);
extern uint64 sys_gettid(void);
extern uint64 sys_prlimit64(void);
extern uint64 sys_clock_gettime(void);
extern uint64 sys_writev(void);
extern uint64 sys_fcntl(void);
extern uint64 sys_mprotect(void);
extern uint64 sys_lseek(void);
extern uint64 sys_futex(void);

extern uint64 sys_socket(void);
extern uint64 sys_socketpair(void);
extern uint64 sys_bind(void);
extern uint64 sys_listen(void);
extern uint64 sys_accept(void);
extern uint64 sys_connect(void);
extern uint64 sys_getsockname(void);
extern uint64 sys_getpeername(void);
extern uint64 sys_sendto(void);
extern uint64 sys_recvfrom(void);
extern uint64 sys_setsockopt(void);
extern uint64 sys_getsockopt(void);

static uint64 (*syscalls[])(void) = {
  [SYS_fork]        sys_fork,
  [SYS_exit]        sys_exit,
  [SYS_wait]        sys_wait,
  [SYS_pipe]        sys_pipe,
  [SYS_read]        sys_read,
  [SYS_exec]        sys_exec,
  [SYS_fstat]       sys_fstat,
  [SYS_chdir]       sys_chdir,
  [SYS_dup]         sys_dup,
  [SYS_getpid]      sys_getpid,
  [SYS_sbrk]        sys_sbrk,
  [SYS_sleep]       sys_sleep,
  [SYS_uptime]      sys_uptime,
  [SYS_open]        sys_open,
  [SYS_write]       sys_write,
  [SYS_mkdir]       sys_mkdir,
  [SYS_close]       sys_close,
  [SYS_test_proc]   sys_test_proc,
  [SYS_dev]         sys_dev,
  [SYS_readdir]     sys_readdir,
  [SYS_getcwd]      sys_getcwd,
  [SYS_remove]      sys_remove,
  [SYS_trace]       sys_trace,
  [SYS_sysinfo]     sys_sysinfo,
  [SYS_rename]      sys_rename,
  [SYS_sprint]      sys_sprint,
  [SYS_openat]      sys_openat,
  [SYS_brk]         sys_brk,
  [SYS_munmap]      sys_munmap,
  [SYS_mmap]        sys_mmap,
  [SYS_getppid]     sys_getppid,
  [SYS_getpid]      sys_getpid,
  [SYS_mkdirat]     sys_mkdirat,
  [SYS_dup3]        sys_dup3,
  [SYS_clone]       sys_clone,
  [SYS_wait4]       sys_wait4,
  [SYS_sched_yield] sys_sched_yield,
  [SYS_times]       sys_times,
  [SYS_nanosleep]   sys_nanosleep,
  [SYS_gettimeofday] sys_gettimeofday,
  [SYS_execve]      sys_execve,
  [SYS_getdents64]  sys_getdents64,
  [SYS_uname]       sys_uname,
  [SYS_unlinkat]    sys_unlinkat,
  [SYS_umount2]     sys_umount,
  [SYS_mount]       sys_mount,
  [SYS_finit]       sys_finit,
  [SYS_writev]      sys_writev,
  [SYS_exit_group]  sys_exit_group,
  [SYS_set_tid_address]  sys_set_tid_address,
  [SYS_clock_gettime]    sys_clock_gettime,
  [SYS_kill]             sys_kill,
  [SYS_rt_sigaction]     sys_rt_sigaction,
  [SYS_rt_sigprocmask]   sys_rt_sigprocmask,
  [SYS_rt_sigtimedwait]  sys_rt_sigtimedwait,
  [SYS_gettid]           sys_gettid,
  [SYS_prlimit64]        sys_prlimit64,
  [SYS_fcntl]           sys_fcntl,
  [SYS_lseek]           sys_lseek,
  [SYS_mprotect]        sys_mprotect,
  [SYS_futex]           sys_futex,
  [SYS_socket]          sys_socket,
  [SYS_socketpair]      sys_socketpair,
  [SYS_bind]            sys_bind,
  [SYS_listen]          sys_listen,
  [SYS_accept]          sys_accept,
  [SYS_connect]         sys_connect,
  [SYS_getsockname]     sys_getsockname,
  [SYS_getpeername]     sys_getpeername,
  [SYS_sendto]          sys_sendto,
  [SYS_recvfrom]        sys_recvfrom,
  [SYS_setsockopt]      sys_setsockopt,
  [SYS_getsockopt]      sys_getsockopt,
};

static char *sysnames[] = {
  [SYS_fork]        "fork",
  [SYS_exit]        "exit",
  [SYS_wait]        "wait",
  [SYS_pipe]        "pipe",
  [SYS_read]        "read",
  [SYS_exec]        "exec",
  [SYS_fstat]       "fstat",
  [SYS_chdir]       "chdir",
  [SYS_dup]         "dup",
  [SYS_getpid]      "getpid",
  [SYS_sbrk]        "sbrk",
  [SYS_sleep]       "sleep",
  [SYS_uptime]      "uptime",
  [SYS_open]        "open",
  [SYS_write]       "write",
  [SYS_mkdir]       "mkdir",
  [SYS_close]       "close",
  [SYS_test_proc]   "test_proc",
  [SYS_dev]         "dev",
  [SYS_readdir]     "readdir",
  [SYS_getcwd]      "getcwd",
  [SYS_remove]      "remove",
  [SYS_trace]       "trace",
  [SYS_sysinfo]     "sysinfo",
  [SYS_rename]      "rename",
  [SYS_sprint]      "sprint",
  [SYS_openat]	    "openat",
  [SYS_brk]         "brk",
  [SYS_munmap]      "munmap",
  [SYS_mmap]        "mmap",
  [SYS_getppid]     "getppid",
  [SYS_getpid]      "getpid",
  [SYS_mkdirat]     "mkdirat",
  [SYS_dup3]        "dup3",
  [SYS_clone]       "clone",
  [SYS_wait4]       "wait4",
  [SYS_sched_yield] "sched_yield",
  [SYS_times]       "times",
  [SYS_nanosleep]   "nanosleep",
  [SYS_gettimeofday] "gettimeofday",
  [SYS_execve]      "execve",
  [SYS_getdents64]  "getdents64",
  [SYS_uname]       "uname",
  [SYS_unlinkat]    "unlinkat",
  [SYS_umount2]     "umount",
  [SYS_mount]       "mount",
  [SYS_finit]       "finit",
  [SYS_writev]       "writev",
  [SYS_exit_group]       "exit_group",
  [SYS_set_tid_address]  "set_tid_address",
  [SYS_clock_gettime]    "clock_gettime",
  [SYS_kill]             "kill",
  [SYS_rt_sigaction]     "sigaction",
  [SYS_rt_sigprocmask]   "sigprocmask",
  [SYS_rt_sigtimedwait]  "sigtimedwait",
  [SYS_gettid]           "gettid",
  [SYS_prlimit64]        "prlimit64",
  [SYS_fcntl]           "fcntl",
  [SYS_mprotect]        "mprotect",
  [SYS_lseek]           "lseek",
  [SYS_futex]           "futex",
  [SYS_socket]          "socket",
  [SYS_socketpair]      "socketpair",
  [SYS_bind]            "bind",
  [SYS_listen]          "listen",
  [SYS_accept]          "accept",
  [SYS_connect]         "connect",
  [SYS_getsockname]     "getsockname",
  [SYS_getpeername]     "getpeername",
  [SYS_sendto]          "sendto",
  [SYS_recvfrom]        "recvfrom",
  [SYS_setsockopt]      "setsockopt",
  [SYS_getsockopt]      "getsockopt",
};

void
syscall(void)
{
  int num;
  struct proc *p = myproc();

  num = p->trapframe->a7;
  if(num > 0 && num < NELEM(syscalls) && syscalls[num]) {
    //printf("syscall %d\n",num);
    //if(num!=66)printf("syscall %d\n",num);
    p->trapframe->a0 = syscalls[num]();
        // trace
    if ((p->tmask & (1 << num)) != 0) {
      printf("pid %d: %s -> %d\n", p->pid, sysnames[num], p->trapframe->a0);
    }
  } else {
    printf("pid %d %s: unknown sys call %d\n",
            p->pid, p->name, num);
    p->trapframe->a0 = -1;
  }
}

uint64 
sys_test_proc(void) {
    int n;
    argint(0, &n);
    printf("hello world from proc %d, hart %d, arg %d\n", myproc()->pid, r_tp(), n);
    return 0;
}

uint64
sys_sysinfo(void)
{
  uint64 addr;
  // struct proc *p = myproc();

  if (argaddr(0, &addr) < 0) {
    return -1;
  }

  struct sysinfo info;
  info.freemem = freemem_amount();
  info.nproc = procnum();

  // if (copyout(p->pagetable, addr, (char *)&info, sizeof(info)) < 0) {
  if (copyout2(addr, (char *)&info, sizeof(info)) < 0) {
    return -1;
  }

  return 0;
}
