//
// File-system system calls.
// Mostly argument checking, since we don't trust
// user code, and calls into file.c and fs.c.
//


#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/stat.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/kalloc.h"
#include "include/sleeplock.h"
#include "include/file.h"
#include "include/pipe.h"
#include "include/fcntl.h"
#include "include/fat32.h"
#include "include/syscall.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/vm.h"
#include "include/mmap.h"

// Fetch the nth word-sized system call argument as a file descriptor
// and return both the descriptor and the corresponding struct file.
extern char dev_img_start[];
extern long dev_img_size;
extern char sacrifice_start[];
extern long sacrifice_size;

#define FILE_ZERO 114514
static int
argfd(int n, int *pfd, struct file **pf)
{
  int fd;
  struct file *f;
  if(argint(n, &fd) < 0)
    return -1;
  if(pfd)
    *pfd = fd;
  if(fd < 0 || fd >= NOFILE || (f=myproc()->ofile[fd]) == NULL)
    return -1;
  if(pf)
    *pf = f;
  return 0;
}

// Allocate a file descriptor for the given file.
// Takes over file reference from caller on success.
static int
fdalloc(struct file *f)
{
  int fd;
  struct proc *p = myproc();

  for(fd = 0; fd < NOFILE; fd++){
    if(p->ofile[fd] == 0){
      p->ofile[fd] = f;
      return fd;
    }
  }
  return -1;
}

uint64
sys_dup(void)
{
  struct file *f;
  int fd;

  if(argfd(0, 0, &f) < 0)
    return -1;
  if((fd=fdalloc(f)) < 0)
    return -1;
  filedup(f);
  return fd;
}

uint64
sys_dup3(void)
{
  struct file *f;
  int newfd;
  
  if(argfd(0, 0, &f) < 0) 
    return -1;
  if(argint(1, &newfd) < 0 || newfd < 0 || newfd >= NOFILE)
    return -1;
  if(myproc()->ofile[newfd] != f) 
  {
    myproc()->ofile[newfd] = f;
    filedup(f);
  }
  return newfd;
}

uint64
sys_read(void)
{
  struct file *f;
  int n;
  uint64 p;
  int zero = 0;
  if(argint(0,&n) >= 0 && argaddr(1, &p) >= 0){
    if(n==FILE_ZERO){
      copyout2(p,(char*)&zero,1);
      return 1;
    }
  }
  if(argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0)
    return -1;
  return fileread(f, p, n);
}

uint64
sys_write(void)
{
  struct file *f;
  int n;
  uint64 p;
  if(argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0){
    return -1;
  }
  
  return filewrite(f, p, n);
}

uint64
sys_close(void)
{
  int fd;
  struct file *f;

  if(argfd(0, &fd, &f) < 0)
    return -1;
  myproc()->ofile[fd] = 0;
  fileclose(f);
  return 0;
}

uint64
sys_fstat(void)
{
  struct file *f;
  int fd;
  uint64 st; // user pointer to struct stat

  if(argfd(0, &fd, &f) < 0 || argaddr(1, &st) < 0)
    return -1;
  //return filestat(f, st);
  return filekstat(f, st);
}

static void
get_parent_name(char *path, char *pname, char *name)
{
  int len = strlen(path);
  strncpy(pname, path, len + 1);
  int i = len - 1;

  if(pname[i] == '/')
  {
    i--;
  }

  for(; i >= 0; --i)
  {
    if(pname[i] == '/')
    {
      pname[i] = 0;
      break;
    }
  }
  int len2 = strlen(pname);
  strncpy(name, path + len2 + 1, len - len2 + 1);
}

static struct dirent*
create(char *path, short type, int mode)
{
  struct dirent *ep, *dp;
  char name[FAT32_MAX_FILENAME + 1];
  char pname[FAT32_MAX_FILENAME + 1];
  
  if (type == T_DIR) {
    mode = ATTR_DIRECTORY;
  } else if (mode & O_RDONLY) {
    mode = ATTR_READ_ONLY;
  } else {
    mode = 0;  
  }

  if((dp = enameparent(path, name)) == NULL)
  {
    get_parent_name(path, pname, name);
    dp = create(pname, T_DIR, O_RDWR);
    if(dp == NULL)
    {
      return NULL;
    } 
  }
  else
  {
    elock(dp);
  }
  
  if ((ep = ealloc(dp, name, mode)) == NULL) {
    eunlock(dp);
    eput(dp);
    return NULL;
  }
  
  if ((type == T_DIR && !(ep->attribute & ATTR_DIRECTORY)) ||
      (type == T_FILE && (ep->attribute & ATTR_DIRECTORY))) {
    eunlock(dp);
    eput(ep);
    eput(dp);
    return NULL;
  }

  eunlock(dp);
  eput(dp);

  elock(ep);
  return ep;
}

static struct dirent*
create_env(struct dirent* env, char *path, short type, int mode)
{
  struct dirent *ep, *dp;
  char name[FAT32_MAX_FILENAME + 1];
  char pname[FAT32_MAX_FILENAME + 1];

  if (type == T_DIR) {
    mode = ATTR_DIRECTORY;
  } else if (mode & O_RDONLY) {
    mode = ATTR_READ_ONLY;
  } else {
    mode = 0;  
  }

  if((dp = enameparent_env(env, path, name)) == NULL)
  {
    get_parent_name(path, pname, name);
    dp = create_env(env, pname, T_DIR, O_RDWR);
    if(dp == NULL)
    {
      return NULL;
    } 
  }
  else
  {
    elock(dp);
  }
  
  if ((ep = ealloc(dp, name, mode)) == NULL) {
    eunlock(dp);
    eput(dp);
    return NULL;
  }
  
  if ((type == T_DIR && !(ep->attribute & ATTR_DIRECTORY)) ||
      (type == T_FILE && (ep->attribute & ATTR_DIRECTORY))) {
    eunlock(dp);
    eput(ep);
    eput(dp);
    return NULL;
  }

  eunlock(dp);
  eput(dp);

  elock(ep);
  return ep;
}

uint64
sys_open(void)
{
  char path[FAT32_MAX_PATH];
  int fd, omode;
  struct file *f;
  struct dirent *ep;
  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argint(1, &omode) < 0)
    return -1;

  if(omode & O_CREATE){
    ep = create(path, T_FILE, omode);
    if(ep == NULL){
      return -1;
    }
  } else {
    if((ep = ename(path)) == NULL){
      return -1;
    }
    elock(ep);
    if((ep->attribute & ATTR_DIRECTORY) && omode != O_RDONLY){
      eunlock(ep);
      eput(ep);
      return -1;
    }
  }

  if((f = filealloc()) == NULL || (fd = fdalloc(f)) < 0){
    if (f) {
      fileclose(f);
    }
    eunlock(ep);
    eput(ep);
    return -1;
  }

  if(!(ep->attribute & ATTR_DIRECTORY) && (omode & O_TRUNC)){
    etrunc(ep);
  }

  f->type = FD_ENTRY;
  f->off = (omode & O_APPEND) ? ep->file_size : 0;
  f->ep = ep;
  f->readable = !(omode & O_WRONLY);
  f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

  eunlock(ep);

  return fd;
}

uint64
sys_mkdir(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = create(path, T_DIR, 0)) == 0){
    return -1;
  }
  eunlock(ep);
  eput(ep);
  return 0;
}

uint64
sys_chdir(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  struct proc *p = myproc();
  
  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = ename(path)) == NULL){
    return -1;
  }
  elock(ep);
  if(!(ep->attribute & ATTR_DIRECTORY)){
    eunlock(ep);
    eput(ep);
    return -1;
  }
  eunlock(ep);
  eput(p->cwd);
  p->cwd = ep;
  return 0;
}

uint64
sys_pipe(void)
{
  uint64 fdarray; // user pointer to array of two integers
  struct file *rf, *wf;
  int fd0, fd1;
  struct proc *p = myproc();

  if(argaddr(0, &fdarray) < 0)
    return -1;
  if(pipealloc(&rf, &wf) < 0)
    return -1;
  fd0 = -1;
  if((fd0 = fdalloc(rf)) < 0 || (fd1 = fdalloc(wf)) < 0){
    if(fd0 >= 0)
      p->ofile[fd0] = 0;
    fileclose(rf);
    fileclose(wf);
    return -1;
  }
  // if(copyout(p->pagetable, fdarray, (char*)&fd0, sizeof(fd0)) < 0 ||
  //    copyout(p->pagetable, fdarray+sizeof(fd0), (char *)&fd1, sizeof(fd1)) < 0){
  if(copyout2(fdarray, (char*)&fd0, sizeof(fd0)) < 0 ||
     copyout2(fdarray+sizeof(fd0), (char *)&fd1, sizeof(fd1)) < 0){
    p->ofile[fd0] = 0;
    p->ofile[fd1] = 0;
    fileclose(rf);
    fileclose(wf);
    return -1;
  }
  return 0;
}

// To open console device.
uint64
sys_dev(void)
{
  int fd, omode;
  int major, minor;
  struct file *f;

  if(argint(0, &omode) < 0 || argint(1, &major) < 0 || argint(2, &minor) < 0){
    return -1;
  }

  if(omode & O_CREATE){
    panic("dev file on FAT");
  }

  if(major < 0 || major >= NDEV)
    return -1;

  if((f = filealloc()) == NULL || (fd = fdalloc(f)) < 0){
    if(f)
      fileclose(f);
    return -1;
  }

  f->type = FD_DEVICE;
  f->off = 0;
  f->ep = 0;
  f->major = major;
  f->readable = !(omode & O_WRONLY);
  f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

  return fd;
}

// To support ls command
uint64
sys_readdir(void)
{
  struct file *f;
  uint64 p;

  if(argfd(0, 0, &f) < 0 || argaddr(1, &p) < 0)
    return -1;
  return dirnext(f, p);
}

// get absolute cwd string
uint64
sys_getcwd(void)
{
  uint64 addr;
  if (argaddr(0, &addr) < 0)
    return -1;

  struct dirent *de = myproc()->cwd;
  char path[FAT32_MAX_PATH];
  char *s;
  int len;

  if (de->parent == NULL) {
    s = "/";
  } else {
    s = path + FAT32_MAX_PATH - 1;
    *s = '\0';
    while (de->parent) {
      len = strlen(de->filename);
      s -= len;
      if (s <= path)          // can't reach root "/"
        return -1;
      strncpy(s, de->filename, len);
      *--s = '/';
      de = de->parent;
    }
  }
  if(addr==0){
    addr=(uint64)kalloc();
    mappages(myproc()->pagetable,addr,PGSIZE,addr,PTE_R|PTE_W);
  }
  // if (copyout(myproc()->pagetable, addr, s, strlen(s) + 1) < 0)
  if (copyout2(addr, s, strlen(s) + 1) < 0)
    return -1;
  return addr;
}

// Is the directory dp empty except for "." and ".." ?
static int
isdirempty(struct dirent *dp)
{
  struct dirent ep;
  int count;
  int ret;
  ep.valid = 0;
  ret = enext(dp, &ep, 2 * 32, &count);   // skip the "." and ".."
  return ret == -1;
}

uint64
sys_remove(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  int len;
  if((len = argstr(0, path, FAT32_MAX_PATH)) <= 0)
    return -1;

  char *s = path + len - 1;
  while (s >= path && *s == '/') {
    s--;
  }
  if (s >= path && *s == '.' && (s == path || *--s == '/')) {
    return -1;
  }
  
  if((ep = ename(path)) == NULL){
    return -1;
  }
  elock(ep);
  if((ep->attribute & ATTR_DIRECTORY) && !isdirempty(ep)){
      eunlock(ep);
      eput(ep);
      return -1;
  }
  elock(ep->parent);      // Will this lead to deadlock?
  eremove(ep);
  eunlock(ep->parent);
  eunlock(ep);
  eput(ep);

  return 0;
}

// Must hold too many locks at a time! It's possible to raise a deadlock.
// Because this op takes some steps, we can't promise
uint64
sys_rename(void)
{
  char old[FAT32_MAX_PATH], new[FAT32_MAX_PATH];
  if (argstr(0, old, FAT32_MAX_PATH) < 0 || argstr(1, new, FAT32_MAX_PATH) < 0) {
      return -1;
  }

  struct dirent *src = NULL, *dst = NULL, *pdst = NULL;
  int srclock = 0;
  char *name;
  if ((src = ename(old)) == NULL || (pdst = enameparent(new, old)) == NULL
      || (name = formatname(old)) == NULL) {
    goto fail;          // src doesn't exist || dst parent doesn't exist || illegal new name
  }
  for (struct dirent *ep = pdst; ep != NULL; ep = ep->parent) {
    if (ep == src) {    // In what universe can we move a directory into its child?
      goto fail;
    }
  }

  uint off;
  elock(src);     // must hold child's lock before acquiring parent's, because we do so in other similar cases
  srclock = 1;
  elock(pdst);
  dst = dirlookup(pdst, name, &off);
  if (dst != NULL) {
    eunlock(pdst);
    if (src == dst) {
      goto fail;
    } else if (src->attribute & dst->attribute & ATTR_DIRECTORY) {
      elock(dst);
      if (!isdirempty(dst)) {    // it's ok to overwrite an empty dir
        eunlock(dst);
        goto fail;
      }
      elock(pdst);
    } else {                    // src is not a dir || dst exists and is not an dir
      goto fail;
    }
  }

  if (dst) {
    eremove(dst);
    eunlock(dst);
  }
  memmove(src->filename, name, FAT32_MAX_FILENAME);
  emake(pdst, src, off);
  if (src->parent != pdst) {
    eunlock(pdst);
    elock(src->parent);
  }
  eremove(src);
  eunlock(src->parent);
  struct dirent *psrc = src->parent;  // src must not be root, or it won't pass the for-loop test
  src->parent = edup(pdst);
  src->off = off;
  src->valid = 1;
  eunlock(src);

  eput(psrc);
  if (dst) {
    eput(dst);
  }
  eput(pdst);
  eput(src);

  return 0;

fail:
  if (srclock)
    eunlock(src);
  if (dst)
    eput(dst);
  if (pdst)
    eput(pdst);
  if (src)
    eput(src);
  return -1;
}

uint64 
sys_openat(void){
  char path[FAT32_MAX_PATH];
  int fd, omode,fd0,mode,env;
  struct file *f;
  struct file *f0;
  struct dirent *dp;
  struct dirent *ep;
  env=0;
  if(argint(0,&fd0) < 0 || argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &omode) < 0
  ||argint(3, &mode)<0)
    return -1;
  if(strncmp(path,"/dev/zero",9)==0){
    return FILE_ZERO;
  }

  if(mode == O_RDWR){
  	omode = omode;
  }else if(mode == 0600){
  	omode = omode;
  	//clear execute
  }
  
  if(path[0] != '/' && fd0 != AT_FDCWD){
    env=1;
    f0 = myproc()->ofile[fd0];
    if(f0 == 0){
      printf("fd:%d illegal\n",fd0);
      return -1;
    }
    dp=f0->ep;
  }

  if(omode & O_CREATE){
    ep = create(path, T_FILE, omode);
    if(ep == NULL){
      printf("create file %s failed\n",path);
      return -1;
    }
  } else {
    if(env){
     if((ep = ename_env(dp,path)) == NULL){
        printf("env path %s not found\n",path);
        return -1;
      }
    }else{
      if((ep = ename(path)) == NULL){
        printf("path %s not found\n",path);
        return -1;
      }
    }
    elock(ep);
    if((ep->attribute & ATTR_DIRECTORY) && (omode != O_RDONLY && omode != O_DIRECTORY)){
      printf("diretory only can be read\n");
      eunlock(ep);
      eput(ep);
      return -1;
    }
  }

  if((f = filealloc()) == NULL || (fd = fdalloc(f)) < 0){
    if (f) {
      fileclose(f);
    }
    printf("file alloc failed\n");
    eunlock(ep);
    eput(ep);
    return -1;
  }

  if(!(ep->attribute & ATTR_DIRECTORY) && (omode & O_TRUNC)){
    etrunc(ep);
  }

  f->type = FD_ENTRY;
  f->off = (omode & O_APPEND) ? ep->file_size : 0;
  f->ep = ep;
  f->readable = !(omode & O_WRONLY);
  f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

  eunlock(ep);
  return fd;

}

uint64
sys_mkdirat(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  struct file *fp;
  int dirfd;
  int mode, omode = 0;

  if(argint(0, &dirfd) < 0 || argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &mode) < 0){
    return -1;
  }
  if(mode == 0666){
      omode = omode;
      // mode 
    }

  if(path[0] == '/' || dirfd == AT_FDCWD) {
    ep = create(path, T_DIR, omode);
  }else {
    fp = myproc()->ofile[dirfd];
    if(!fp) return -1;
    ep = create_env(fp->ep, path, T_DIR, omode);
  }

  eunlock(ep);
  eput(ep);
  return 0;
}


uint64
sys_mmap(void)
{
  uint64 start;
  uint64 len;
  int prot;
  int flags;
  int fd;
  int off;
  if(argaddr(0, &start) < 0)
    return -1;
  if(argaddr(1, &len) < 0)
    return -1;
  if(argint(2, &prot) < 0)
    return -1;
  if(argint(3, &flags) < 0)
    return -1;
  if(argfd(4, &fd, NULL) < 0 && fd!=-1){
    //printf("fd:%d\n",fd);
    return -1;
  }
  if(argint(5, &off) < 0)
    return -1;

  return do_mmap(start, len, prot, flags, fd, off);
}

uint64
sys_munmap(void)
{
  uint64 start;
  uint64 len;
  if(argaddr(0, &start) < 0 || argaddr(1, &len) < 0){
    return -1;
  }
  return do_munmap(start, len,1);
}


uint64
sys_getdents64(void) 
{
  struct file* fp;
  int fd;
  uint64 buf;
  uint64 len;

  if(argfd(0, &fd, &fp) < 0 || argaddr(1, &buf) < 0 || argaddr(2, &len) < 0) {
    return -1;
  }

  return dirent_next(fp, buf, len);
}

uint64
sys_unlinkat(void){
  struct file* fp;
  int dirfd,flags,len;
  struct dirent *dp;
  struct dirent* ep;
  char path[FAT32_MAX_PATH];
  
  if(argint(0,&dirfd)<0||
     (len = argstr(1,path,FAT32_MAX_PATH))<0||
     argint(2,&flags)<0){
    return -1;   
  }
  
  if(path[0] != '/' && dirfd != AT_FDCWD){
    fp = myproc()->ofile[dirfd];
    if(fp == NULL)return -1;
    dp = fp->ep;
  }else
    dp = &rootfs->root;
  
  char *s = path + len - 1;
  while (s >= path && *s == '/') {
    s--;
  }
  if (s >= path && *s == '.' && (s == path || *--s == '/')) {
    return -1;
  }
  
  if((ep = ename_env(dp,path)) == NULL){
    return -1;
  }
  elock(ep);
  if((ep->attribute & ATTR_DIRECTORY) && !isdirempty(ep)){
      eunlock(ep);
      eput(ep);
      return -1;
  }
  elock(ep->parent);      // Will this lead to deadlock?
  eremove(ep);
  eunlock(ep->parent);
  eunlock(ep);
  eput(ep);
  return 0;
}


uint64 
sys_umount(void)
{
  char mntpoint[FAT32_MAX_PATH];
  if(argstr(0, mntpoint, FAT32_MAX_PATH)<0) {
    return -1;
  }
  eumount(mntpoint);
  return 0;
}

uint64 
sys_mount(void)
{
  char device[FAT32_MAX_PATH];
  char mntpoint[FAT32_MAX_PATH];
  char fs_type[20];
  if((argstr(0, device, FAT32_MAX_PATH)<0)||
     (argstr(1, mntpoint, FAT32_MAX_PATH)<0)||
     (argstr(2, fs_type, 20)<0)) {
    return -1;
  }
  if(strncmp(fs_type,"vfat",20)==0){
    struct fs* img_fs = NULL;
    struct dirent* img = NULL;
    img = ename(device); 
    if(!img)return -1;
    img_fs = fat32_img(img);
    if(!img_fs)return -1;
    if(emount(img_fs,mntpoint)<0)
    	return -1;
    return 0;
  }else
    return -1;
}

uint64 
sys_writev(void){
  return 0;
}

uint64
sys_fcntl(void){
  int fd;
  argint(0, &fd);
  if(fd >= SOCKET) return 0xffffffff;
  else return 0;
}

uint64
sys_mprotect(void){
  return 0;
}

uint64
sys_finit(void){
  struct dirent *dp;
  struct dirent *ep;
  printf("finit\n");
  dp = create("/dev", T_DIR, 0);
  eunlock(dp);
  ep = create_env(dp,"/vda2",T_FILE,0);
  ewrite(ep, 0, (uint64)dev_img_start, 0, dev_img_size);
  eunlock(ep);
  eput(dp);
  eput(ep);
  
  ep = create("/sacrifice",T_FILE,0);
  ewrite(ep, 0, (uint64)sacrifice_start, 0, sacrifice_size);
  eunlock(ep);
  eput(ep);

  ep = create("/etc/passwd", T_FILE, 0);
  eunlock(ep);
  eput(ep);
  return 0;
}

uint64 sys_lseek(void)
{
  struct file *f;
  uint64 offset;
  int fd;
  int whence;
  uint64 ret = -1;

  if(argfd(0, &fd, &f) < 0 || argaddr(1, &offset) < 0 || argint(2, &whence) < 0)
  {
    return -1;
  }

  ret = filelseek(f, offset, whence);
  return ret;
}

