#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/syscall.h"
#include "include/timer.h"
#include "include/kalloc.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/sbi2.h"
#include "include/console.h"
#include "include/string.h"
#include "include/vm.h"
#include "include/uname.h"

extern struct proc proc[NPROC];
extern int exec(char *path, char **argv);

uint64
sys_exec(void)
{
  char path[FAT32_MAX_PATH], *argv[MAXARG];
  int i;

  uint64 uargv, uarg;
  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argaddr(1, &uargv) < 0){
    return -1;
  }

  memset(argv, 0, sizeof(argv));
  for(i=0;; i++){
    if(i >= NELEM(argv)){
      goto bad;
    }
    if(fetchaddr(uargv+sizeof(uint64)*i, (uint64*)&uarg) < 0){
      goto bad;
    }
    if(uarg == 0){
      argv[i] = 0;
      break;
    }
    argv[i] = kalloc();
    if(argv[i] == 0){
      goto bad;
    }
    if(fetchstr(uarg, argv[i], PGSIZE) < 0){
      goto bad;
    }
  }

  int ret = exec(path, argv);

  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  return ret;

 bad:
  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  printf("sys exec bad\n");
  return -1;
}

uint64
sys_execve(void)
{

  char path[FAT32_MAX_PATH], *argv[MAXARG];
  int i;
  uint64 uargv, uarg;
  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argaddr(1, &uargv) < 0){
    return -1;
  }
  //printf("1\n");
  memset(argv, 0, sizeof(argv));
  for(i=0;; i++){
    if(i >= NELEM(argv)){
      goto bad;
    }
    if(fetchaddr(uargv+sizeof(uint64)*i, (uint64*)&uarg) < 0){
      goto bad;
    }
    if(uarg == 0){
      argv[i] = 0;
      break;
    }
    argv[i] = kalloc();
    if(argv[i] == 0){
      goto bad;
    }
    if(fetchstr(uarg, argv[i], PGSIZE) < 0){
      goto bad;
    }
  }
  
  int ret = exec(path, argv);

  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  return ret;

 bad:
  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  printf("sys exec bad\n");
  return -1;
}

uint64
sys_exit(void)
{
  int n;
  if(argint(0, &n) < 0)
    return -1;
  exit(n);
  return 0;  // not reached
}

uint64
sys_getpid(void)
{
  return myproc()->pid;
}

uint64
sys_fork(void)
{
  return fork();
}

uint64
sys_wait(void)
{
  uint64 p;
  int f;
  if(argaddr(0, &p) < 0)
    return -1;
  f = wait(p);
  return f;
}

uint64
sys_sbrk(void)
{
  int addr;
  int n;

  if(argint(0, &n) < 0)
    return -1;
  addr = myproc()->sz;
  if(growproc(n) < 0)
    return -1;
  return addr;
}

uint64
sys_sleep(void)
{
  int n;
  uint ticks0;

  if(argint(0, &n) < 0)
    return -1;
  acquire(&tickslock);
  ticks0 = ticks;
  while(ticks - ticks0 < n){
    if(myproc()->killed){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;
}

uint64
sys_kill(void)
{
  int pid;

  if(argint(0, &pid) < 0)
    return -1;
  return kill(pid);
}

// return how many clock tick interrupts have occurred
// since start.
uint64
sys_uptime(void)
{
  uint xticks;

  acquire(&tickslock);
  xticks = ticks;
  release(&tickslock);
  return xticks;
}

uint64
sys_trace(void)
{
  int mask;
  if(argint(0, &mask) < 0) {
    return -1;
  }
  myproc()->tmask = mask;
  return 0;
}



uint64
sys_sprint(void){
  char buf[256];
  if(argstr(0, buf,256) < 0)
    return -1;
  printf("%s",buf);
  return 1;
}

uint64
sys_getppid(void){
  struct proc* p =myproc();
  return p->parent->pid;
}

uint64
sys_brk(void)
{
  int n;
  if(argint(0, &n) < 0)
    return -1;
  if(n == 0) n = 0;
  else n = n / 256;
  return growproc(n);
}

uint64 
sys_clone(void) {
	uint64 flag, stack, tls;
	if (argaddr(0, &flag) < 0) 
		return -1;
	if (argaddr(1, &stack) < 0) 
		return -1;
  // printf("flag:%p\n",flag);
  if(argaddr(4,&tls) < 0)
    return -1;
  if(stack != 0 || flag & 0x200000 || flag & 0x1000000){
	  return clone(flag, stack, (int*)tls);
  }
  else
    return fork();
}

uint64
sys_wait4(void)
{
  uint64 p;
  int pid;
  if(argint(0, &pid) < 0)
    return -1;
  if(argaddr(1, &p) < 0)
    return -1;
  if (pid == -1){
    int f = wait(p);
    if(p != 0 && f > 0){
      //char k = 0x00;
      //copyout2(p,(char*)&k,1);
      //k = 0x00;
      //copyout2(p + 1,(char*)&k,1);
    }
    return f;
  }
  else{
    int f = wait4pid(pid, p);
    if(p != 0 && f > 0){
      //int k = 0x0300;
      //copyout2(p,(char*)&k,sizeof(int));
    }
    return f;
  }
  //return wait(p);
}

uint64
sys_sched_yield(){
  yield();
  return 0;
}
/*
//unfinished, just to pass the test.
uint64
sys_times(){
  uint64 p;
  struct tms tms;
  tms.tms_utime = 1;
  tms.tms_stime = 1;
  tms.tms_cutime = 1;
  tms.tms_cstime = 1;   
  if(argaddr(0, &p) < 0)
    return -1;
  copyout2(p,(char*)&tms,sizeof(tms));
  return 1;
}
*/
uint64 sys_nanosleep(void) {
	uint64 addr_sec, addr_usec;

	if (argaddr(0, &addr_sec) < 0) 
		return -1;
	if (argaddr(1, &addr_usec) < 0) 
		return -1;

	struct proc *p = myproc();
	uint64 sec, usec;
	if (copyin2((char*)&sec, addr_sec, sizeof(sec)) < 0) 
		return -1;
	if (copyin2((char*)&usec, addr_usec, sizeof(usec)) < 0) 
		return -1;
	int n = sec * 20 + usec / 50000000;

	int mask = p->tmask;
	if (mask) {
		printf(") ...\n");
	}
	acquire(&p->lock);
	uint64 tick0 = ticks;
	while (ticks - tick0 < n / 10) {
		if (p->killed) {
			return -1;
		}
		sleep(&ticks, &p->lock);
	}
	release(&p->lock);

	return 0;
}

uint64 sys_uname(void) {
	uint64 addr;

	if (argaddr(0, &addr) < 0) {
		return -1;
	}

	return uname_copyout(addr);
}

uint64
sys_exit_group(void){
  //printf("exit group\n");
  return 0;
}

uint64
sys_set_tid_address(void){
  uint64 address;
  argaddr(0, &address);
  // printf("set-tid-address: address:%p\n",address);
  uint64 pa = walkaddr(myproc()->pagetable,address);
  // printf("set-tid-address:pa:%p\n");
  myproc()->clear_child_tid = (int *)pa;
  // copyin2((char*)&myproc()->clear_child_tid, address ,8);
  
  // do_futex(myproc()->clear_child_tid, FUTEX_WAKE, 1, NULL, NULL, 0, 0);
  return myproc()->pid;
}

uint64 sys_rt_sigaction(void) {
	int signum;
	uint64 uptr_act;		// struct sigaction const *act
	uint64 uptr_oldact;		// struct sigaction *oldact
	int size;

	// extract user args
	// if (argint(0, &signum) < 0) {
	// 	return -EINVAL;
	// }
	// if (argaddr(1, &uptr_act) < 0) {
	// 	return -EINVAL;
	// }
	// if (argaddr(2, &uptr_oldact) < 0) {
	// 	return -EINVAL;
	// }
	argint(0, &signum);
	argaddr(1, &uptr_act);
	argaddr(2, &uptr_oldact);
	argint(3, &size);

	// copy struct sigaction from user space 
	struct sigaction act;
	struct sigaction oldact;
	// if (uptr_act && copyin2((char*)&act, uptr_act, sizeof(struct sigaction)) < 0) {
	// 	return -EFAULT;
	// }
	if (size > 16 * 8) {
		size = 16 * 8;
	}
	else if (size < 8) {
		size = 8;
	}

	if (uptr_act) {
		if (
			copyin2((char*)&(act.__sigaction_handler), uptr_act, sizeof(__sighandler_t)) < 0 
			// copyin2((char*)&(act.sa_mask), uptr_act + sizeof(__sighandler_t), size) < 0 || 
			// copyin2((char*)&(act.sa_flags), uptr_act + sizeof(__sighandler_t) + size, sizeof(int)) < 0
		) {
			return -1;
		}
	}

	if (set_sigaction(
		signum, 
		uptr_act ? &act : NULL, 
		uptr_oldact ? &oldact : NULL, 
		size / 8
	) < 0) {
		return -1;
	}

	// copyout old struct sigaction 
	// if (uptr_oldact && copyout2(uptr_oldact, (char*)&oldact, sizeof(struct sigaction)) < 0) {
	// 	return -EFAULT;
	// }
	if (uptr_oldact) {
		if (
			copyout2(uptr_oldact, (char*)&(act.__sigaction_handler), sizeof(__sighandler_t)) < 0 
			// copyout2(uptr_oldact + sizeof(__sighandler_t), (char*)&(act.sa_mask), size) < 0 || 
			// copyout2(uptr_oldact + sizeof(__sighandler_t) + size, (char*)&(act.sa_flags), sizeof(int)) < 0
		) {
			return -1;
		}
	}

	return 0;
}

uint64 sys_rt_sigprocmask(void) {
	int how;
	uint64 uptr_set, uptr_oldset;

	__sigset_t set, oldset;
	int size;

	argint(0, &how);
	argaddr(1, &uptr_set);
	argaddr(2, &uptr_oldset);
	argint(3, &size);

	if (size > 16 * 8) {
		size = 16 * 8;
	}
	else if (size < 8) {
		size = 8;
	}

	// copy in __sigset_t 
	if (uptr_set && copyin2((char*)&set, uptr_set, size) < 0) {
		return -1;
	}

	if (sigprocmask(
		how, 
		&set, 
		uptr_oldset ? &oldset : NULL, 
		size / 8
	)) {
		return -1;
	}

	if (uptr_oldset && copyout2(uptr_oldset, (char*)&oldset, size) < 0) {
		return -1;
	}

	return 0;
}


uint64
sys_rt_sigtimedwait(void){
  //printf("sigtimedwait\n");
  return 0;
}

uint64
sys_gettid(void){
  //temporily think it's a proc...
  return myproc()->pid;
  //printf("sigtimedwait\n");
  //return 0;
}

// k210 says it's no important to complete.
uint64
sys_prlimit64(void){
  //printf("prlimit64\n");
  return 0;
}

uint64
sys_futex(void){
  // printf("Into futex");
  uint64 uaddr;
  int op;
  int val;
  uint64 utime;
  struct timespec ut;
  ktime_t kt;
  uint64 uaddr2;
  int val2 = 0;
  int val3;
  if (argaddr(0, &uaddr) < 0) {
		return -1;
	}
  if (argint(1, &op) < 0) {
		return -1;
	}
  if (argint(2, &val) < 0) {
		return -1;
	}
  if (argaddr(3, &utime) < 0)
  {
    return -1;
  }
  if (argaddr(4, &uaddr2) < 0)
  {
    return -1;
  }
  if (argint(5, &val3) < 0)
  {
    return -1;
  }

  int cmd = op & FUTEX_CMD_MASK;
  // printf("addr:%p,op:%d,val:%d,cmd=%d,uaddr:%p\n",uaddr,op,val,cmd);
  if(utime == 0)
  {
    // printf("no timeout\n");
  }

  if(cmd == FUTEX_WAIT)
  {
    copyout2(utime, (char*)&ut, sizeof(struct timespec));
    kt = ktime_set(ut.tv_sec, ut.tv_nsec);
  }

  if (cmd == FUTEX_REQUEUE || cmd == FUTEX_CMP_REQUEUE ||
	    cmd == FUTEX_CMP_REQUEUE_PI || cmd == FUTEX_WAKE_OP)
		val2 = (uint32) (unsigned long) utime;

  do_futex((int*)uaddr, op, val, &kt, (int *)uaddr2, val2, val3);
  return 0;
}